/******************************************************************/
/* igluVideoCapture.cpp                                           */
/* -----------------------                                        */
/*                                                                */
/*                                                                */
/* Chris Wyman (10/06/2011)                                       */
/******************************************************************/

#if defined(HAS_FFMPEG)
#define __STDC_FORMAT_MACROS
extern "C" {
#include <avcodec.h>
#include <avformat.h>
#include <swscale.h>
}
#endif

#include <cstdio>

#pragma warning( disable: 4996 )

#include "iglu.h"

using namespace iglu;

char *g_formatNames[IGLU_FORMATS_MAX] = {
	"avi", "flv", "h263", "h264", "ipod", "mp4", "mjpeg", "mpeg", "psp" 
};

char *g_codecNames[IGLU_CODECS_MAX] = {
	"libxvid", "mpeg4", "msmpeg4", "libx264", "flv", "h263", "mjpeg", "mpeg2video", "wmv2", "wmv3"	
};

#if defined(HAS_FFMPEG)
CodecID g_codecIDs[IGLU_CODECS_MAX] = {
	CODEC_ID_MPEG4, CODEC_ID_MPEG4, CODEC_ID_MSMPEG4V3, CODEC_ID_H264, CODEC_ID_FLASHSV, CODEC_ID_H263I, CODEC_ID_MJPEG, CODEC_ID_MPEG2VIDEO, CODEC_ID_WMV2, CODEC_ID_WMV3
};
#endif

// A little helper class useful only internally in this video implementation.
class iglu::IGLUVideoIOData
{
public:

#if defined(HAS_FFMPEG)
	IGLUVideoIOData() : encoderSwsContext(0), videoStream(-1) {}

	AVFormatContext *pFormatCtx;
	AVCodecContext  *pCodecCtx;
	AVCodec         *pCodec;
	AVFrame         *pFrame; 
    AVFrame         *pFrameRGB;
	SwsContext      *encoderSwsContext;
	AVOutputFormat  *pOutputFormat;
	int              videoStream;
	AVPacket        *pPacket;
#else
	IGLUVideoIOData() {}
#endif
};


void IGLUVideoCapture::ListSupportedFormatsAndCodecs( void )
{
#if defined(HAS_FFMPEG)
	AVOutputFormat *fmt = 0;
	while (fmt = av_oformat_next(fmt)) 
	{
         if (fmt->name && fmt->long_name)
			 printf("Supported format:  %s (aka '%s')\n", fmt->long_name, fmt->name );
    }

	AVCodec *cd = 0;
	while (cd = av_codec_next(cd)) 
	{
         if (cd->name && (cd->type == AVMEDIA_TYPE_VIDEO) )
			 printf("Supported video codec:  %s (%s)\n", cd->long_name, cd->name );
    }
#endif
}

IGLUVideoCapture::IGLUVideoCapture( ) :
	m_isOpen(false), m_curFrame(0), m_width(-1), m_height(-1), m_bitrate(400000), m_framerate(30)
{
#if defined(HAS_FFMPEG)
	// Make sure FFMpeg is set up.
	avcodec_register_all();
    av_register_all();

	// Surpress annoying and amazingly verbose FFmpeg log messages to stderr
	av_log_set_level( AV_LOG_QUIET );
	
	// Create our internal data structure to store FFMpeg state
	m_vidData = new IGLUVideoIOData();

#else
	ErrorExit( "IGLU compiled without FFMpeg support.  Unable to use IGLUVideoCapture()!", 
		       __FILE__, __FUNCTION__, __LINE__ );
#endif
}

bool IGLUVideoCapture::StartCapture( char* filename, int width, int height, int codec, int bitrate, int format, int frameRate )
{
#if defined(HAS_FFMPEG)
	m_width = width;
	m_height = height;
	m_bitrate = bitrate;
	m_framerate = frameRate;

	// Identify the specified codec, and try to initialize it!
	m_vidData->pOutputFormat = av_guess_format( g_formatNames[format], NULL, NULL );
	if (!m_vidData->pOutputFormat)
	{
		Warning( "IGLU unable to locate specified FFMpeg video codec!", __FILE__, __FUNCTION__, __LINE__ );
		return false;
	}

	// Create a context
	m_vidData->pFormatCtx = avformat_alloc_context();
	if (!m_vidData->pFormatCtx)
	{
		Warning( "IGLU unable to allocate FFMpeg video context!", __FILE__, __FUNCTION__, __LINE__ );
		return false;
	}
	m_vidData->pFormatCtx->oformat = m_vidData->pOutputFormat;
	sprintf( m_vidData->pFormatCtx->filename, "%s", filename );

	// Find our encoder
	m_vidData->pCodec = avcodec_find_encoder( g_codecIDs[codec] );
	if (!m_vidData->pCodec)
	{
		Warning( "IGLU unable to locate appropriate FFMpeg encoder!", __FILE__, __FUNCTION__, __LINE__ );
		return false;
	}

	// Get and initialize our codec context
	m_vidData->pCodecCtx = avcodec_alloc_context3( m_vidData->pCodec );
	m_vidData->pCodecCtx->codec_id       = g_codecIDs[codec]; // m_vidData->pOutputFormat->video_codec;
	m_vidData->pCodecCtx->codec_type     = AVMEDIA_TYPE_VIDEO;
	m_vidData->pCodecCtx->gop_size       = 10;   // Look on Wikipedia for MPEG-2.  Size of "Group of Pictures"
	m_vidData->pCodecCtx->pix_fmt        = PIX_FMT_YUV420P;
	m_vidData->pCodecCtx->width          = width;
	m_vidData->pCodecCtx->height         = height;
	m_vidData->pCodecCtx->bit_rate       = bitrate;
	m_vidData->pCodecCtx->time_base.den  = frameRate;
	m_vidData->pCodecCtx->time_base.num  = 1;

	if ( m_vidData->pOutputFormat->flags & AVFMT_GLOBALHEADER )
		m_vidData->pCodecCtx->flags |= CODEC_FLAG_GLOBAL_HEADER;

	// Open up our codec
	if (avcodec_open2( m_vidData->pCodecCtx, m_vidData->pCodec, NULL ) < 0)
	{
		Warning( "IGLU unable to open FFMpeg codec!", __FILE__, __FUNCTION__, __LINE__ );
		return false;
	}

	// Create a temporary buffer for frame output data
	m_tmpFrameBufSize = m_width*m_height*3;
	m_tmpFrameBuf = (unsigned char *)malloc( sizeof(uint8_t) * m_tmpFrameBufSize );
	if (!m_tmpFrameBuf)
	{
		Warning( "IGLU unable to allocation temporary buffer for video encoding!", __FILE__, __FUNCTION__, __LINE__ );
		return false; 
	}

	if (!InitializeVideoFrame())
		return false;

	// Open the video file
	if ( url_fopen( &m_vidData->pFormatCtx->pb, filename, URL_WRONLY ) < 0 )
	{
		Warning("Unable to open video file for writing!", __FILE__, __FUNCTION__, __LINE__ );
		return false;
	}

	// Write any video header this particular format/codec needs.
	av_write_header( m_vidData->pFormatCtx );

#endif

	return (m_isOpen = true);
}

bool IGLUVideoCapture::FinishCapture( void )
{
#if defined(HAS_FFMPEG)

	if (!m_isOpen) return true;

	printf("writing trailer\n");
	av_write_trailer( m_vidData->pFormatCtx );
	printf("closing video file\n");
	url_fclose( m_vidData->pFormatCtx->pb );
	printf("closing codec\n");
	avcodec_close( m_vidData->pCodecCtx );

	// Freeing temporary buffers
	printf("freeing temporary buffers\n");
	free( m_tmpFrameBuf );
	av_free( m_vidData->pFormatCtx );
	av_free( m_vidData->pCodecCtx );

	if (m_imgData) free( m_imgData );
	if (m_vidData->pFrame) av_free( m_vidData->pFrame );
	if (m_imgData) free( m_imgDataRGB );
	if (m_vidData->pFrame) av_free( m_vidData->pFrameRGB );

	m_isOpen = false;

#endif

	return true;
}

bool IGLUVideoCapture::InitializeVideoFrame( void )
{
#if defined(HAS_FFMPEG)
	// Allocate a temporary frame for our video
	m_vidData->pFrame = avcodec_alloc_frame();
	m_vidData->pFrameRGB = avcodec_alloc_frame();
	if (!m_vidData->pFrame || !m_vidData->pFrameRGB) return false;

	// Allocate a buffer to store data in this frame
	int numBytes = avpicture_get_size( m_vidData->pCodecCtx->pix_fmt, m_vidData->pCodecCtx->width, m_vidData->pCodecCtx->height );
	m_imgData = (unsigned char *)malloc( numBytes * sizeof( unsigned char ) );
	if (!m_imgData) {
		av_free( m_vidData->pFrame );
		return false;
	}
	numBytes = avpicture_get_size( PIX_FMT_RGB24, m_vidData->pCodecCtx->width, m_vidData->pCodecCtx->height );
	m_imgDataRGB = (unsigned char *)malloc( numBytes * sizeof( unsigned char ) );
	if (!m_imgDataRGB) {
		av_free( m_vidData->pFrameRGB );
		return false;
	}

	// Assign appropriate parts of buffer to image planes in the frames
    avpicture_fill((AVPicture *)m_vidData->pFrame, 
		           (uint8_t *)m_imgData, 
				   m_vidData->pCodecCtx->pix_fmt,
                   m_vidData->pCodecCtx->width, 
				   m_vidData->pCodecCtx->height);
	avpicture_fill((AVPicture *)m_vidData->pFrameRGB, 
		           (uint8_t *)m_imgDataRGB, 
				   PIX_FMT_RGB24,
                   m_vidData->pCodecCtx->width, 
				   m_vidData->pCodecCtx->height);
#endif
	return true;
}

bool IGLUVideoCapture::CaptureFrame( unsigned char *frame, int width, int height )
{
	if (!m_isOpen) 
	{
		Warning("CaptureFrame() called without first calling StartCapture()!", __FILE__, __FUNCTION__, __LINE__ );
		return false;
	}
	if (width != m_width || height != m_height) 
	{
		printf("Initial size: (%d %d).  Current frame: (%d %d)\n", m_width, m_height, width, height );
		return false;
	}

#if defined(HAS_FFMPEG)
	// Copy frame RGB data into internal structure, unless the user already put it there!
	if (frame != 0)
		memcpy( m_imgDataRGB, frame, width*height*3*sizeof(unsigned char) );

	// Convert RGB -> YUV so we can encode the frame
	m_vidData->encoderSwsContext = sws_getCachedContext( m_vidData->encoderSwsContext,
			                                      m_vidData->pCodecCtx->width, m_vidData->pCodecCtx->height, 
												  PIX_FMT_RGB24, 
												  m_vidData->pCodecCtx->width, m_vidData->pCodecCtx->height, 
												  m_vidData->pCodecCtx->pix_fmt,
												  SWS_BICUBIC, NULL, NULL, NULL );
	sws_scale( m_vidData->encoderSwsContext, m_vidData->pFrameRGB->data, m_vidData->pFrameRGB->linesize, 0, 
					                          m_vidData->pCodecCtx->height, m_vidData->pFrame->data, 
											  m_vidData->pFrame->linesize );

	// Go ahead and encode the frame
	int gotPacket = 0;
	int out_size = avcodec_encode_video( m_vidData->pCodecCtx, (uint8_t *)m_tmpFrameBuf, m_tmpFrameBufSize, m_vidData->pFrame );
	printf("Frame size: %d\n",out_size);
	if (out_size > 0)
	{
		av_init_packet( m_vidData->pPacket );
		if (m_vidData->pCodecCtx->coded_frame->pts != AV_NOPTS_VALUE)
			m_vidData->pPacket->pts = av_rescale_q( m_vidData->pPacket->pts, m_vidData->pCodecCtx->time_base, m_vidData->pCodecCtx->time_base );
		if (m_vidData->pCodecCtx->coded_frame->key_frame)
			m_vidData->pPacket->flags |= AV_PKT_FLAG_KEY;
		m_vidData->pPacket->stream_index = 0;
		m_vidData->pPacket->data = m_tmpFrameBuf;
		m_vidData->pPacket->size = out_size;
		if( av_interleaved_write_frame( m_vidData->pFormatCtx, m_vidData->pPacket ) < 0 )
			return false;
	}
	
#endif

	return true;
}


