/******************************************************************/
/* igluFilePath.cpp                                               */
/* -----------------------                                        */
/*                                                                */
/* A header for interfacing with a simplistic filepath manager    */
/*     and an IGLU version of fopen() for loading files from      */
/*     specified path directories.                                */
/*                                                                */
/* Chris Wyman (10/20/2011)                                       */
/******************************************************************/

#include "iglu.h"
#include <sys/types.h>
#include <sys/stat.h>

using namespace iglu;

// Annonyous namespace storing globals -- an IGLU built-in set of file paths
namespace {
	IGLUFilePath igluPath;
};

void iglu::AddSearchPath( const char *dirName, int pathType )
{
	igluPath.AddPath( dirName, pathType );
}

bool iglu::FileExitsInPath( const char *fileName, int pathType )
{
	return igluPath.FileExists( fileName, pathType ) != 0;
}

FILE *iglu::FileOpen( const char *fileName, const char *fileMode, int pathType )
{
	return igluPath.FileOpen( fileName, fileMode, pathType );
}

char *iglu::FindFileInPath( const char *fileName, int pathType )
{
	return igluPath.FileInPath( fileName, pathType );
}


IGLUFilePath::IGLUFilePath() 
{
	m_searchPaths[COMMON].Add( strdup( "" ) );
}

IGLUFilePath::~IGLUFilePath() 
{
	for (int j=0; j<MAX_PATH_TYPES; j++)
		for (uint i=0; i<m_searchPaths[j].Size(); i++)
			free( m_searchPaths[j][i] );
}


void IGLUFilePath::AddPath( const char *dirName, int pathType )
{
	m_searchPaths[pathType].Add( strdup( dirName ) );
}

bool IGLUFilePath::Exists( const char *file )
{
	// Apparently the function stat() is system-dependent.  For Windows
	//    use Microsoft's _stat().  On other systems, try glibc's stat() version.
	// XXX: This is not currently tested beyond Windows...
#if defined(WIN32) && defined(_MSC_VER)
	struct _stat tmp;
	return (_stat( file, &tmp ) == 0);
#else
	struct stat tmp;
	return (stat( file, &tmp ) == 0);
#endif
}

int IGLUFilePath::FileExists( const char *fileName, int pathType )
{
	// Check all the paths in the specified search path
	for (uint i=0; i<m_searchPaths[pathType].Size(); i++)
	{
		sprintf( m_currentFile, "%s%s", m_searchPaths[pathType][i], fileName );
		if ( Exists( m_currentFile ) )
			return (i+1);
	}

	// If the specified path was NOT the common path, also look in the common path
	if (pathType!=COMMON)
		return -1*FileExists( fileName, COMMON );

	return 0;
}

FILE *IGLUFilePath::FileOpen( const char *fileName, const char * fileMode, int pathType )
{
	// Check if the file exists (and if so, where)
	int exists = FileExists( fileName, pathType );

	// File doesn't exist
	if (!exists) return 0;

	// Open and return
	return std::fopen( m_currentFile, fileMode );
}

char *IGLUFilePath::FileInPath( const char *fileName, int pathType )
{
	// Check if the file exists (and if so, where)
	int exists = FileExists( fileName, pathType );

	// File doesn't exist
	if (!exists) return 0;

	// Open and return
	return m_currentFile;
}