#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <math.h>
#include "igluPFM.h"

extern void FatalError( char *msg );
extern void FatalError( char *formatStr, char *insertThisStr );
extern void Error( char *msg );
extern void Error( char *formatStr, char *insertThisStr );
extern void Warning( char *msg );
extern void Warning( char *formatStr, char *insertThisStr );



float *iglu::ReadPFM( char *f, int *mode, int *width, int *height, float *scale )
{
	FILE *infile;
	char line[256];
	float endianScale, *image;
	int curMode;
	long img_size, count;
	bool valid=true;
    
	/* double check that the pointers for req'd output parameters are non-null */
	if (!width || !height)
		FatalError( "ReadPFM() called with null width/height parameters!" );

	/* open file containing texture */
	if ((infile = fopen(f, "rb")) == NULL) 
		FatalError( "Fatal Error: IGLU: Can't open file '%s'!", f );
  
	/* read and parse first line (the Pf, PF, or PF4 format specifier) */
	fgets(line, 256, infile);
	valid = (line[0] != 'P' && line[0] != 'p') ? false : valid;
	valid = (line[1] != 'F' && line[1] != 'f') ? false : valid;
	valid = (line[2] == '4' && line[1] == 'f') ? false : valid;
	if (!valid)
		FatalError( "Fatal Error: IGLU: Invalid PFM format in '%s'!", f );
	curMode = (line[1] == 'f') ? PFM_GRAY : PFM_RGB;
	curMode = (line[2] == '4') ? PFM_RGBA : curMode;
	if (mode) *mode = curMode;

	/* discard all the comments at the top of the file               */
	/* if there are comments elsewhere, this code doesn't handle it! */
	fgets(line, 256, infile);
	while (line[0] == '#')
		fgets(line, 256, infile);

	/* read image size and max component value */
	sscanf(line, "%d %d", width, height);
  
	/* read endianness and float scale */
	fscanf(infile, "%f ", &endianScale);
    if (endianScale > 0)
		FatalError( "Fatal Error: IGLU done not handle big-endian PFMs! (%s)", f );
	endianScale = fabs( endianScale );
	if (scale) *scale = endianScale;

	/* allocate texture array */
	img_size  = (*height)*(*width);
	img_size *= (curMode == PFM_RGB)  ? 3 : 1;
	img_size *= (curMode == PFM_RGBA) ? 4 : 1;
	if ((image = (float *)calloc(img_size, sizeof(float))) == NULL)
		FatalError( "IGLU: Cannot allocate memory for image! (%s)", f );
  
	/* read the image data */
	count = fread( image, sizeof(float), img_size, infile );
  
	/* make input data size matches expected size! */
	if (count != img_size)
		Warning( "IGLU: Header and file data sizes do not agree! (%s)", f );

	/* close the input files */
	fclose( infile );
  
	/* done! */
	return image;
}




int iglu::WritePFM( char *f, int mode, int width, int height, float *ptr, float scale )
{
	time_t thetime;
	long size;
	FILE *out;

	/* double check the format type is OK */
	if ( mode<PFM_GRAY || mode>PFM_RGBA ) 
	{
		Error( "IGLU: Bad PFM format type!" );
		return GFXIO_UNSUPPORTED;
	}

	/* open the specified file */
	out = fopen(f, "wb");
	if (!out) {
		char buf[512];
		sprintf( buf, "IGLU: Unable to open file '%s', output lost!", f );
		Error( buf );
		return GFXIO_OPENERROR;
	}
  
	/* write out the appropriate format specifier */
	if (mode==PFM_GRAY)
		fprintf(out, "Pf\n");  
	else if (mode==PFM_RGB)
		fprintf(out, "PF\n");
	else 
		fprintf(out, "PF4\n");

	/* add a comment specifying when the image was created */
	thetime = time(0);
	fprintf(out, "# File created by Chris Wyman's PFM Library on %s",
			ctime(&thetime));

	/* write the rest of the header */
	fprintf(out, "%d %d\n", width, height);
    fprintf(out, "%f\n", -1*scale);  // -1 because we're little endian....  +1 if big endian
  
	/* how much stuff will we write out? */
	size = width*height;
	size *= (mode==PFM_RGB ) ? 3 : 1;
	size *= (mode==PFM_RGBA) ? 4 : 1;

	/* actually write out the data */
	fwrite( ptr, sizeof(float), size, out );

	/* clean up and return */
	fclose(out);
	return GFXIO_OK;
}