/**********************************
** igluPNG.cpp                   **
** -----                         **
**                               **
** Code to read PNG files.  This **
**   relies on IGLU linking to   **
**   FLTK, which has a built-in  **
**   PNG class.                  **
**                               **
** Chris Wyman (1/21/2013)       **
**********************************/

#include <cstdio>
#include <cstdlib>
#include <cstring>
#include "igluPNG.h"
#include "errors/igluErrorExit.h"
#include "png/png.h"

using namespace iglu;


// The code for reading PNGs below thanks to Morten Nobel (and, of course, libpng)

unsigned char *iglu::ReadPNG( char *f, int *width, int *height, int *components, bool invertY ) 
{
    png_structp png_ptr;
    png_infop info_ptr;
    unsigned int sig_read = 0;
    FILE *fp;

    if ((fp = fopen(f, "rb")) == NULL)
        return false;

    png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING,
            NULL, NULL, NULL);

    if (png_ptr == NULL) {
        fclose(fp);
        return false;
    }

    /* Allocate/initialize the memory
     * for image information.  REQUIRED. */
    info_ptr = png_create_info_struct(png_ptr);
    if (info_ptr == NULL) {
        fclose(fp);
        png_destroy_read_struct(&png_ptr, NULL, NULL);
        return false;
    }

    /* Set error handling if you are
     * using the setjmp/longjmp method
     * (this is the normal method of
     * doing things with libpng).
     * REQUIRED unless you  set up
     * your own error handlers in
     * the png_create_read_struct()
     * earlier.
     */
    if (setjmp(png_jmpbuf(png_ptr))) {
        /* Free all of the memory associated
         * with the png_ptr and info_ptr */
        png_destroy_read_struct(&png_ptr, &info_ptr, NULL);
        fclose(fp);
        /* If we get here, we had a
         * problem reading the file */
        return false;
    }

    /* Set up the output control if
     * you are using standard C streams */
    png_init_io(png_ptr, fp);

    /* If we have already
     * read some of the signature */
    png_set_sig_bytes(png_ptr, sig_read);

    png_read_png(png_ptr, info_ptr, PNG_TRANSFORM_STRIP_16 | PNG_TRANSFORM_PACKING | PNG_TRANSFORM_EXPAND, NULL);

	png_uint_32 tmpWidth, tmpHeight;
	int bitDepth, colorType;
	png_get_IHDR(png_ptr, info_ptr, &tmpWidth, &tmpHeight, &bitDepth, &colorType, NULL, NULL, NULL);

    *width = tmpWidth;
    *height = tmpHeight;
    switch (colorType) {
        case PNG_COLOR_TYPE_RGBA:
            *components = 4;
            break;
        case PNG_COLOR_TYPE_RGB:
            *components = 3;
            break;
        default:
			printf("*** Error! PNG image type not currently handled! (%s)\n", f);
            png_destroy_read_struct(&png_ptr, &info_ptr, NULL);
            fclose(fp);
            return false;
    }
    unsigned int row_bytes = png_get_rowbytes(png_ptr, info_ptr);
    unsigned char *tmpData = (unsigned char*) malloc(row_bytes * tmpHeight);

    png_bytepp row_pointers = png_get_rows(png_ptr, info_ptr);

    for (unsigned int i = 0; i < tmpHeight; i++) {
        // Copy this inverted or not, dependant in function inputs
		memcpy( tmpData +(row_bytes * (invertY?(tmpHeight-1-i):i)), 
			    row_pointers[i], row_bytes);
    }

    /* Clean up after the read,
     * and free any memory allocated */
    png_destroy_read_struct(&png_ptr, &info_ptr, NULL);

    /* Close the file */
    fclose(fp);

    /* That's it */
    return tmpData;
}



