/************************************************************
** igluTGA.cpp                                             **
** -----                                                   **
**                                                         **
** Code to read Targa/TGA files.  This only reads the most **
**   basic of TGA files (e.g., no compression or fancy     **
**   formats).                                             **
**                                                         **
** Chris Wyman (1/21/2013)                                 **
************************************************************/

#include <cstdio>
#include <cstdlib>
#include <cstring>
#include "igluTGA.h"
#include "errors/igluErrorExit.h"

using namespace iglu;

void FatalError( char *msg );
void FatalError( char *formatStr, char *insertThisStr );

namespace {
	typedef struct {
		int status;
		unsigned char type, pixelDepth;
		short int width, height;
		unsigned char *imageData;
	} tgaInfo;
};




// load the image header fields. We only keep those that matter!
void tgaLoadHeader(FILE *file, tgaInfo *info) {

	unsigned char cGarbage;
	short int iGarbage;

	fread(&cGarbage, sizeof(unsigned char), 1, file);
	fread(&cGarbage, sizeof(unsigned char), 1, file);

// type must be 2 or 3
	fread(&info->type, sizeof(unsigned char), 1, file);

	fread(&iGarbage, sizeof(short int), 1, file);
	fread(&iGarbage, sizeof(short int), 1, file);
	fread(&cGarbage, sizeof(unsigned char), 1, file);
	fread(&iGarbage, sizeof(short int), 1, file);
	fread(&iGarbage, sizeof(short int), 1, file);

	fread(&info->width, sizeof(short int), 1, file);
	fread(&info->height, sizeof(short int), 1, file);
	fread(&info->pixelDepth, sizeof(unsigned char), 1, file);

	fread(&cGarbage, sizeof(unsigned char), 1, file);
}

// loads the image pixels. You shouldn't call this function
// directly
void tgaLoadImageData(FILE *file, tgaInfo *info) {

	int mode,total,i;
	unsigned char aux;

// mode equal the number of components for each pixel
	mode = info->pixelDepth / 8;
// total is the number of bytes we'll have to read
	total = info->height * info->width * mode;
	
	fread(info->imageData,sizeof(unsigned char),total,file);

// mode=3 or 4 implies that the image is RGB(A). However TGA
// stores it as BGR(A) so we'll have to swap R and B.
	if (mode >= 3)
		for (i=0; i < total; i+= mode) {
			aux = info->imageData[i];
			info->imageData[i] = info->imageData[i+2];
			info->imageData[i+2] = aux;
		}
}	

// this is the function to call when we want to load
// an image
unsigned char *iglu::ReadTGA( char *f, int *width, int *height, int *components ) 
{
	FILE *file;
	tgaInfo *info;
	int mode,total;
	unsigned char *data;

	info = (tgaInfo *)malloc(sizeof(tgaInfo));
	if (!info) FatalError( "Memory allocation error in ReadTGA()!" );

	file = fopen(f, "rb");
	if (!file) LoadError( f, "ReadTGA()" );

	tgaLoadHeader(file,info);
	if (ferror(file)) FatalError( "Error while reading '%s'\n", f );
	if (info->type == 1) FatalError( "ReadTGA() found '%s' contains indexed colors.\n", f );
	if (info->type != 2) FatalError( "ReadTGA() found '%s' contains non-RGB or compressed image.\n", f );
	
	mode  = info->pixelDepth / 8;
	total = info->height * info->width * mode;
	data  = info->imageData = (unsigned char *)malloc(sizeof(unsigned char) * total);
	if (!data) FatalError( "Memory allocation error in ReadTGA()!" );

	*width      = info->width;
	*height     = info->height;
	*components = info->pixelDepth / 8;

	tgaLoadImageData(file,info);
	if (ferror(file)) FatalError( "Error while reading '%s'\n", f );
	fclose(file);
	if (info) free(info);

	return data;
}		


