/******************************************************************/
/* igluBinaryObjArrayReader.cpp                                   */
/* -----------------------                                        */
/*                                                                */
/* Defines a reader for our custom binary object files and their  */
/*     associated (.obj-style) materials                          */
/*                                                                */
/* Chris Wyman (10/20/2011)                                       */
/******************************************************************/

#include "iglu.h"
#include <float.h>

using namespace iglu;


IGLUBinaryObjArrayReader::IGLUBinaryObjArrayReader( char *filename, int params ) :
	IGLUModel(), m_vertArr(0), m_hasMatlID(true), m_hasObjectID(true),
	m_hasTexCoords(false), m_hasNormals(false), m_hasVertices(false), m_shaderID(0)
{
	// Create the data structure to interface with OpenGL for drawing this object.
	m_vertArr = new IGLUVertexArray();

	// Parse the parameters the binary reader pays attention to
	m_loadMtlFile = params & IGLU_OBJ_NO_MATERIALS ? false : true;

	if (params & IGLU_OBJ_VERBOSE_READ)
		printf("(+) Loading %s...\n", filename);

	// Go ahead and load the binary blob.
	LoadBinaryBlob( filename );

	// Read any material files attached to this blob
	if (m_loadMtlFile)
		for (uint i=0; i<m_objMtlFiles.Size(); i++)
			IGLUOBJMaterialReader reader( m_objMtlFiles[i] );

	// OK.  Now we've created our vertex array.  Enable the attributes so that we're sending the correct
	//    data to all our shaders.
	m_vertArr->EnableAttribute( IGLU_ATTRIB_VERTEX, 3, GL_FLOAT, m_vertStride, BUFFER_OFFSET(m_vertOff));
	if (HasNormals())  
		m_vertArr->EnableAttribute( IGLU_ATTRIB_NORMAL, 3, GL_FLOAT, m_vertStride, BUFFER_OFFSET(m_normOff));
	if (HasTexCoords())   
		m_vertArr->EnableAttribute( IGLU_ATTRIB_TEXCOORD, 2, GL_FLOAT, m_vertStride, BUFFER_OFFSET(m_texOff));
	if (HasMatlID() && m_loadMtlFile)  
		m_vertArr->EnableAttribute( IGLU_ATTRIB_MATL_ID, 1, GL_FLOAT, m_vertStride, BUFFER_OFFSET(m_matlIdOff));
	if (HasObjectID()) 
		m_vertArr->EnableAttribute( IGLU_ATTRIB_OBJECT_ID, 1, GL_FLOAT, m_vertStride, BUFFER_OFFSET(m_objectIdOff));

	if (params & IGLU_OBJ_VERBOSE_READ)
		printf("   (-) Loaded %d vertices, %d triangles: %d\n", m_numVertices, m_numElems);
}

IGLUBinaryObjArrayReader::~IGLUBinaryObjArrayReader()
{
	// Free strings contining the mtllib filenames
	for (uint i=0; i<m_objMtlFiles.Size(); i++)
		free( m_objMtlFiles[i] );

	// Get rid of our vertex array
	delete m_vertArr;
}

bool IGLUBinaryObjArrayReader::LoadBinaryBlob( char *file )
{
	// Open the binary blob file
	FILE *f = iglu::FileOpen( file, "rb", IGLUFilePath::MODEL );
	if (!f)	
		LoadError( file, "IGLUBinaryOBJArrayReader()" );

	// Read the header
	IGLUBinaryVBOHeader hdr;
	fread( &hdr, sizeof(IGLUBinaryVBOHeader), 1, f );

	// Double check that this is a version that we understand...
	if (hdr.vboVersionID != 1)
		LoadError( file, "IGLUBinaryOBJArrayReader::Version()" );

	// Populate class data based on header information
	m_numVertices  = hdr.numVerts;
	m_numElems     = hdr.numElems;
	m_primType     = hdr.elemType;
	m_vertStride   = hdr.vertStride;
	m_vertOff      = hdr.vertOffset;
	m_matlIdOff    = hdr.matlOffset;
	m_objectIdOff  = hdr.objOffset;
	m_normOff      = hdr.normOffset;
	m_texOff       = hdr.texOffset;
	if (hdr.matlFileName[0] != 0)
		m_objMtlFiles.Add( strdup( hdr.matlFileName ) );
	m_hasVertices  = ( hdr.vertBitfield & HAS_VERTEX ) > 0;
	m_hasNormals   = ( hdr.vertBitfield & HAS_NORMAL ) > 0;
	m_hasTexCoords = ( hdr.vertBitfield & HAS_TEXCOORD ) > 0;
	m_hasMatlID    = ( hdr.vertBitfield & HAS_MATLID ) > 0;
	m_hasObjectID  = ( hdr.vertBitfield & HAS_OBJID ) > 0;

	// Compute the sizes of the vertex and element arrays we need to read
	int vertArrSz  = m_numVertices * m_vertStride;
	int elemArrSz  = m_numElems * sizeof( int );

	// Load our vertex array, send it down to the GPU
	void *vPtr = malloc( vertArrSz );
	fread( vPtr, vertArrSz, 1, f );
	m_vertArr->SetVertexArray( vertArrSz, vPtr, IGLU_STATIC|IGLU_DRAW );
	free( vPtr );
	
	// Load our element array, send it down to the GPU
	void *ePtr = malloc( elemArrSz );
	fread( ePtr, elemArrSz, 1, f );
	m_vertArr->SetElementArray( GL_UNSIGNED_INT, elemArrSz, ePtr, IGLU_STATIC|IGLU_DRAW );
	free( ePtr );

	// Done reading from our file!
	fclose( f );

	return true;
}



int IGLUBinaryObjArrayReader::Draw( void )
{
	m_vertArr->DrawElements( m_primType, m_numElems );
	return IGLU_NO_ERROR;
}

int IGLUBinaryObjArrayReader::Draw( IGLUShaderProgram::Ptr &shader )
{
	// If the shader in question is not already enabled, enable it!
	bool wasShaderBound = shader->IsEnabled();
	if (!wasShaderBound)
		shader->Enable();

	m_vertArr->DrawElements( m_primType, m_numElems );

	// If we started with the shader disabled, disable it again now.
	if (!wasShaderBound)
		shader->Disable();

	return IGLU_NO_ERROR;
}