/******************************************************************/
/* igluImage.cpp                                                  */
/* -----------------------                                        */
/*                                                                */
/* The file defines an image class that reads images of various   */
/*     kinds into a uniform internal structure.                   */
/*                                                                */
/* Chris Wyman (10/03/2011)                                       */
/******************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "iglu.h"
#include "iglu/igluImage.h"
#include "Images/igluPPM.h"
#include "Images/igluBMP.h"
#include "Images/igluRGB.h"
#include "Images/igluJPEG.h"
#include "Images/igluPFM.h"
#include "Images/igluPNG.h"
#include "Images/igluTGA.h"
#include <GL/glew.h>

using namespace iglu;

IGLUImage::IGLUImage( char *filename ) :
	m_imgData(0), m_width(-1), m_height(-1),
	m_glDatatype( GL_UNSIGNED_BYTE ), m_glFormat( GL_RGB8 ),
	m_freeMemory(true), m_numComponents(-1)
{
	m_qualifiedFileName = strdup( FindFileInPath( filename, IGLUFilePath::IMAGE ) );
	if (!m_qualifiedFileName)
		LoadError( filename, "IGLUImage" );

	// Identify the type of file
	char *ptr = strrchr( filename, '.' );
	char buf[16];
	strncpy( buf, ptr, 16 );
	for (int i=0;i<16;i++) buf[i] = tolower( buf[i] );

	// Load the image, if possible.
	bool success = false;

	// Is it a PPM file?
	if (!strcmp(buf, ".ppm") || !strcmp(buf, ".pgm"))	
		success = LoadPPM( m_qualifiedFileName );

	// Is it a PFM file?
	else if (!strcmp(buf, ".pfm") || !strcmp(buf, ".pf4"))	
		success = LoadPFM( m_qualifiedFileName );
	
	// Is it a RGB/SGI file?
	else if (!strcmp(buf, ".rgb") || !strcmp(buf, ".rgba"))
		success = LoadRGB( m_qualifiedFileName );
	
	// Is it a BMP file?
	else if (!strcmp(buf, ".bmp"))
		success = LoadBMP( m_qualifiedFileName );

	// Is it a JPEG file?
	else if (!strcmp(buf, ".jpg") || !strcmp(buf, ".jpeg"))
		success = LoadJPEG( m_qualifiedFileName );

	// Is it a PNG file?
	else if (!strcmp(buf, ".png"))
		success = LoadPNG( m_qualifiedFileName );

	// Is it a TGA/Targa file?
	else if (!strcmp(buf, ".tga"))
		success = LoadTGA( m_qualifiedFileName );

	// We don't know how to load this file...
	else
		printf("IGLUImage:  Unknown file format for '%s'...\n", filename);

	// Did we fail to load the image for one reason or another?
	if (!success)
		m_width = m_height = -1;
}

IGLUImage::IGLUImage( unsigned char* image, int width, int height, bool useAlpha, bool freeMemory ) :
    m_width( width ), m_height( height ), m_glFormat( useAlpha ? GL_RGBA8 : GL_RGB8 ),
	m_glDatatype( GL_UNSIGNED_BYTE ), m_freeMemory(freeMemory), m_qualifiedFileName(0),
	m_numComponents( useAlpha ? 4 : 3 )
{
	m_imgData = image;
}

IGLUImage::~IGLUImage()
{
	if (m_qualifiedFileName) free( m_qualifiedFileName );
	if (m_imgData && m_freeMemory) free(m_imgData);
}


bool IGLUImage::LoadRGB( char *filename )
{
	m_imgData = ReadRGB(filename, &m_width, &m_height, &m_numComponents, true);
	m_glFormat = GL_RGBA8;  // ReadRGB always returns 4 bytes per pixel, even if numCompnents < 4.
	m_glDatatype = GL_UNSIGNED_BYTE;

	if (!m_imgData) 
		return false;
	return true;
}


bool IGLUImage::LoadPPM( char *filename )
{
	int mode;
	m_imgData = ReadPPM( filename, &mode, &m_width, &m_height, false );
	m_glFormat = GL_RGB8;
	m_numComponents = 3;
	m_glDatatype = GL_UNSIGNED_BYTE;

	if (!m_imgData) 
		return false;
	return true;
}

bool IGLUImage::LoadPFM ( char *filename )
{
	int mode;
	float scale;
	m_imgData = (unsigned char *)ReadPFM( filename, &mode, &m_width, &m_height, &scale );
	m_glFormat = GL_RGB32F;  m_numComponents = 3;
	m_glFormat = (mode==PFM_GRAY) ? GL_R32F  : m_glFormat;
	m_glFormat = (mode==PFM_RGBA) ? GL_RGBA32F : m_glFormat;
	m_glDatatype = GL_FLOAT;
	m_numComponents = 3;
	m_numComponents = (mode==PFM_GRAY) ? 1 : m_numComponents;
	m_numComponents = (mode==PFM_RGBA) ? 4 : m_numComponents;

	if (!m_imgData) 
		return false;
	return true;
}


bool IGLUImage::LoadBMP( char *filename )
{
	m_imgData = ReadBMP( filename, &m_width, &m_height, true );
	m_glFormat = GL_RGB8;
	m_numComponents = 3;
	m_glDatatype = GL_UNSIGNED_BYTE;

	if (!m_imgData) 
		return false;
	return true;
}

bool IGLUImage::LoadJPEG( char *filename )
{
	m_imgData = ReadJPEG( filename, &m_width, &m_height );
	m_glFormat = GL_RGB8;
	m_numComponents = 3;
	m_glDatatype = GL_UNSIGNED_BYTE;

	if (!m_imgData) 
		return false;
	return true;
}

bool IGLUImage::LoadPNG( char *filename )
{
	m_imgData = ReadPNG( filename, &m_width, &m_height, &m_numComponents );
	m_glFormat = (m_numComponents == 3 ? GL_RGB8 : GL_RGBA8);
	m_glDatatype = GL_UNSIGNED_BYTE;

	if (!m_imgData) 
		return false;
	return true;
}

bool IGLUImage::LoadTGA ( char *filename )
{
	m_imgData = ReadTGA( filename, &m_width, &m_height, &m_numComponents );
	m_glFormat = (m_numComponents == 3 ? GL_RGB8 : GL_RGBA8);
	m_glDatatype = GL_UNSIGNED_BYTE;

	if (!m_imgData) 
		return false;
	return true;
}


void IGLUImage::SaveAsPPM( char *ppmFile )
{
	if (!IsValid()) return;

	if (m_glFormat != GL_RGB8 && m_glFormat != GL_RGBA8)
	{
		printf("Error! IGLUImage::SaveAsPPM() called on image with more than 8-bits/channel!\n");
		exit(-1);
	}

	WritePPM( ppmFile, PPM_RAW, m_width, m_height, m_imgData );
}


