/******************************************************************/
/* igluRandomTexture3D.cpp                                        */
/* -----------------------                                        */
/*                                                                */
/* The file defines an image class for passing a 3D array of      */
/*    random unsigned intergers down to a GLSL program.  This     */
/*    is based of the internal IGLU random number generator to    */
/*    create random seeds.                                        */
/*                                                                */
/* Note: When using the random seeds on the GPU, ensure you use   */
/*    a random number generator that uses *different* adder and   */
/*    multipliers than the built-in IGLU generator, otherwise you */
/*    may get strange correlation artifacts between adjacent      */
/*    pixels!                                                     */
/*                                                                */
/* Chris Wyman (12/07/2011)                                       */
/******************************************************************/

#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <cctype>
#include <GL/glew.h>
#include "iglu.h"

using namespace iglu;


IGLURandomTexture3D::IGLURandomTexture3D( uint width, uint height, uint depth, GLenum textureType,
										  unsigned int randomSeed, unsigned int flags, bool initializeImmediately ):
	IGLUTexture(), m_rndSeed(randomSeed), m_randType(textureType)
{
	m_minFilter     = GL_NEAREST; 
	m_magFilter     = GL_NEAREST; 
	m_sWrap         = GetSWrap( flags );
	m_tWrap         = GetTWrap( flags );
	m_uWrap         = GL_CLAMP_TO_EDGE;
	m_mipmapsNeeded = false; 
	m_compressTex   = false;

	m_width         = width;
	m_height        = height;
	m_depth         = depth;

	m_rng           = new IGLURandom( m_rndSeed );

	if ( (m_randType != GL_FLOAT) && (m_randType != GL_INT) )
		ErrorExit( "Invalid type specified to IGLURandomTexture3D()!", __FILE__, __FUNCTION__, __LINE__ );

	// Ok, the user wants us to initialize the OpenGL texture immediately (not wait until
	//    they later call Initialize() separately).
	if (initializeImmediately)
	{
		CheckValidOpenGL( __FILE__, __FUNCTION__, __LINE__ );
		Initialize();
	}
}

IGLURandomTexture3D::~IGLURandomTexture3D()
{
	if (m_rng) delete m_rng;
}

void IGLURandomTexture3D::Initialize( void )
{
	// Make sure to avoid doubly-calling initialize
	if (m_initialized) return;

	// OK, now we can set up our texture
	glBindTexture( GL_TEXTURE_3D, m_texID );
	glTexParameteri( GL_TEXTURE_3D, GL_TEXTURE_MIN_FILTER, m_minFilter );
	glTexParameteri( GL_TEXTURE_3D, GL_TEXTURE_MAG_FILTER, m_magFilter );
	glTexParameteri( GL_TEXTURE_3D, GL_TEXTURE_WRAP_S, m_sWrap );
	glTexParameteri( GL_TEXTURE_3D, GL_TEXTURE_WRAP_T, m_tWrap );
	glTexParameteri( GL_TEXTURE_3D, GL_TEXTURE_WRAP_R, m_uWrap );

	if ( m_randType == GL_FLOAT )
	{
		float *rands = new float[ m_width * m_height * m_depth ];
		for (int i=0; i<m_width*m_height*m_depth; i++)
			rands[i] = (float)m_rng->fRandom();
		glTexImage3D( GL_TEXTURE_3D, 0, GL_R32F,  
					  m_width, m_height, m_depth, 0, GL_RED, GL_FLOAT, 
					  (void *)rands );
		delete rands;
	}
	else // if ( m_randType == GL_INT )
	{
		uint *rands = new uint[ m_width * m_height * m_depth ];
		for (int i=0; i<m_width*m_height*m_depth; i++)
			rands[i] = (uint)(65535 * m_rng->fRandom());
		glTexImage3D( GL_TEXTURE_3D, 0, GL_R32I,  
					  m_width, m_height, m_depth, 0, GL_RED_INTEGER, GL_UNSIGNED_INT, 
					  (void *)rands );
		delete rands;
	}

	glBindTexture( GL_TEXTURE_3D, 0 );

	// Done initializing
	m_initialized = true;
}


void IGLURandomTexture3D::Resize(int width, int height, int depth){
	
	//Set new values of height and width 
	m_width         = width;
	m_height        = height;
	m_depth         = depth;
	
	//If already initialized, then initialize with new size.
	//	Otherwise wait for the user to call Initialize()
	if(m_initialized){
		m_initialized = false;
		Initialize();
	}
	
}
