/******************************************************************/
/* igluGLContext.cpp                                              */
/* -----------------------                                        */
/* A class encapsulating basic information and settings for the   */
/*    current OpenGL context.                                     */
/*                                                                */
/*  --> Currently, everything only works for a single GL context  */
/*                                                                */
/* Chris Wyman (06/13/2011)                                       */
/******************************************************************/

#include "iglu.h"

using namespace iglu;

char  IGLUContext::m_glVerStr[256]     = "Current OpenGL Context: (not initialized)";
int   IGLUContext::glMajorVer          = 1;           // Default
int   IGLUContext::glMinorVer          = 0;           // Default
uint  IGLUContext::glProfileBits       = PROFILE_COMPATIBILITY;  
uint  IGLUContext::glFlags             = FLAGS_NONE; 
int   IGLUContext::m_glAttribList[128] = { 0, 0 };

void IGLUContext::SetGLVersion( int majorVer, int minorVer, uint profileBits )
{
	glMajorVer = majorVer;
	glMinorVer = minorVer;
	glProfileBits = profileBits;
}

char *IGLUContext::GetGLVersion( void )
{
	if (!glGetString(GL_VERSION))
	{
		sprintf(m_glVerStr, "Current OpenGL Context: (not initialized)");
		return m_glVerStr;
	}

	// Get information about the context profile
	GLint mask;
	glGetIntegerv( GL_CONTEXT_PROFILE_MASK, &mask );
	char prof[128];
	if      (mask & PROFILE_CORE)          sprintf(prof, "(Profile: Core)");
	else if (mask & PROFILE_COMPATIBILITY) sprintf(prof, "(Profile: Compat)");
	else                                   sprintf(prof, "");

	char debug[128];
	sprintf(debug, "%s", IsDebugContext() ? "[Debug]" : "" );

	sprintf(m_glVerStr, "Current OpenGL Context: %s %s %s\n", glGetString( GL_VERSION ), prof, debug );
	return m_glVerStr;
}

bool  IGLUContext::IsCoreContext( void )
{
	GLint mask;
	glGetIntegerv( GL_CONTEXT_PROFILE_MASK, &mask );
	return (mask & PROFILE_CORE) > 0;
}

bool  IGLUContext::IsCompatibilityContext( void )
{
	GLint mask;
	glGetIntegerv( GL_CONTEXT_PROFILE_MASK, &mask );
	return (mask & PROFILE_COMPATIBILITY) > 0;
}

bool  IGLUContext::IsDebugContext( void )
{
	GLint mask;
	glGetIntegerv( GL_CONTEXT_FLAGS, &mask );
	return (mask & FLAGS_DEBUG) > 0;
}

int  *IGLUContext::GetContextAttribList( void )
{
	// Index into the current position in the array
	int i=0;

	// Create a default, which when used will function equivalently to calling the 
	//    to calling the {wgl,glx,agl}CreateContext() function. 
	m_glAttribList[0] = m_glAttribList[1] = 0;

	// If we're using the default settings, go ahead and use this default attriblist
	if (glMajorVer==1 && glMinorVer==0)
		return m_glAttribList;

	// Ok, the user actually specified a major/minor OpenGL version.  Use that.
	m_glAttribList[i++] = 0x2091;      // {GLX,WGL}_CONTEXT_MAJOR_VERSION_ARB
	m_glAttribList[i++] = glMajorVer;
	m_glAttribList[i++] = 0x2092;      // {GLX,WGL}_CONTEXT_MINOR_VERSION_ARB
	m_glAttribList[i++] = glMinorVer;

	// Do we have any context flags?
	if (glFlags != FLAGS_NONE) 
	{
		m_glAttribList[i++] = 0x2094;      // {GLX,WGL}_CONTEXT_FLAGS_ARB
		m_glAttribList[i++] = glFlags;
	}

	// Do we have any specific profile bits?
	if (glProfileBits > 0)
	{
		m_glAttribList[i++] = 0x9126;      // {GLX,WGL}_CONTEXT_PROFILE_MASK_ARB
		m_glAttribList[i++] = glProfileBits;
	}

	// End our list.  Various documentation either suggest you need one or two
	//   zeros to end the list.  We'll put two, just in case.
	m_glAttribList[i++] = 0;
	m_glAttribList[i++] = 0;

	// Ok!  Done creating our list.  Send it off.
	return m_glAttribList;
}